/*
 * TemperatureMonitor7MQTTClient 
 * Misura la temperatura di N_REMOTE_TEMP_SENS sensori remoti e di dsDeviceCount sensori di temperatura locali
 * e lettura dello stato del bruciatore
*/

#include <Arduino.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <WiFi.h>
#include <WiFiUdp.h> 

#define WIFI_STATUS                 21                              // Blue LED on NINA W102
#define TEMP_SENSx_PIN              18                              // Temperature sensor 1..5 input pin (CMD) era 18
#define ADC_TEMP_RESOLUTION         12

#define TIME_MS                     1000                            // 1000 [ms] = 1s  timeout
#define TEMPERATURE_SAMPLING_TIME_S 10                              // 10 [s] temperature sampling time 
#define MAX_TEMPERATURE_SENSORS     10                              // Max mumber of DS local temperature sensor supported
#define N_REMOTE_TEMP_SENS          1                               // Number of the remote temperature sensor, works only with 1 in this example
#define REMOTE_SENSOR_TIMEOUT_MS    2000                            // [ms] Remote sensor timeout (base time 1s)
#define LOCAL_SENSOR_TIMEOUT_MS     2000                            // [ms] Local sensor timeout  (base time 1s)
#define TX_RETRAY_N                 3                               //  mumber of retransmission to request remote sensor temperature

#define RELOAD_TIMER(T)      (SwTimer##T = TimeOut##T)
#define SET_TIMEOUT(T, VAL)  (TimeOut##T = (VAL)/TIME_MS)
#define TIMEOUT_EVENT(T)     (SwTimer##T == 0)

volatile uint8_t  SwTimer1;
         uint8_t  TimeOut1;

volatile uint8_t  SwTimer2;
         uint8_t  TimeOut2;
 
volatile uint32_t DayTimerSec = 0;                                  // Count the number of seconds in each day. from 0 to 86400-1 second in 1 day

volatile bool     SamplingEvent;                                    // Sampling event of temperature sensors (local and remote)

bool              WiFiConnected = false;
bool              EnableRemoteSensor = true;
                  
// Setup a oneWire instance to communicate with any OneWire devices (not just Maxim/Dallas temperature ICs)
OneWire           dsWire(TEMP_SENSx_PIN);
DallasTemperature ds(&dsWire);
DeviceAddress     ThermometersAdd[MAX_TEMPERATURE_SENSORS];
uint8_t           dsDeviceCount;
unsigned long     lastmils=0;

float             Temperatures[MAX_TEMPERATURE_SENSORS];            // Images of sensor temperatures updated every
                                         
typedef enum {
  STOP,          
  MEASURING,
  COMPLETED,      
}T_MT_STATES;                                                       // Measure temperature states

T_MT_STATES LocalMeasureState = STOP;                               // Status of the local temperatures measuuring process
T_MT_STATES RemoteMeasureState = STOP;                              // Status of the remote temperatures measuuring process

// WiFi network name and password:
const char * networkName = "Vodafone-27586605";
const char * networkPswd = "Mypassword";

//IP address to send UDP data to: either use the ip address of the server or a network broadcast address
const char * udpAddress = "192.168.10.255";
const int    udpPort_base = 1234;                 // base port number for remote sensor. moltiple sensor are accessiblle at udpPort_base, udpPort_base + 1, udpPort_base + N_REMOTE_TEMP_SENS -1
char         packetBuffer[255];

WiFiUDP      Udp;  
uint8_t      retry_cnt = TX_RETRAY_N-1;

// function to print a device address
void printAddress(DeviceAddress deviceAddress)
{
  Serial.print("DS12B80 address: ");
  for (uint8_t i = 0; i < 8; i++)
  {
    // zero pad the address if necessary
    if (deviceAddress[i] < 16) Serial.print("0");
    Serial.print(deviceAddress[i], HEX);
  }
  Serial.println();
}


void Temperatures_print(void)
{ 
  String date_time;
   
    for (int i = 0; i<dsDeviceCount+N_REMOTE_TEMP_SENS; i++)
      Serial.println("TEMP" + String(i+1) + ": " + String(Temperatures[i]) + "°C");
}

T_MT_STATES TemperatureMeasurement_start(void)
{            
    Serial.print("Local temperature conversion start time: ");
    Serial.println(millis());
    ds.requestTemperatures();             // Send the command to start conversion of temperatures    
    RELOAD_TIMER(2);
    LocalMeasureState = MEASURING;
    return(LocalMeasureState);
}

T_MT_STATES isTemperatureMeasurementState(void)
{     
  bool completed = true;      
  float t;
          
    if (LocalMeasureState == MEASURING)
    {            
      // completed = ds.isConversionComplete();      
      
      if (TIMEOUT_EVENT(2))      
      { 
        for (int i = 0; i<dsDeviceCount; i++)
        {          
          t = ds.getTempCByIndex(i);    // bug ! it doesn't return the measure
          //t = ds.getTempCByIndex(i);  // if read 2 time then work, but if the program runs for hours the return value is frozen,
                                        // to a specific value that doesn't change if the temperature change.
          Temperatures[i] = t;
        }  
        
        Serial.print("Local temperature value: ");  
        Serial.println(t);
        Serial.print("Local temperature conversion stop time: ");  
        Serial.println(millis());
        
        LocalMeasureState = COMPLETED;                
                
      } // if completed
    } // if MeasuringTemperature

    return(LocalMeasureState);
}

T_MT_STATES TemperatureMeasurement_stop(void)
{
    // Pay attention, to test if the last measure has been completed use isTemperatureMeasurementState
    // then invoke TemperatureMeasurement_stop;
    LocalMeasureState = STOP;
    return(LocalMeasureState);
}

/* ---------------------------------------------------------------------- Remote Temperature measueerment -------------------------------------------------------- */

T_MT_STATES RemoteSensorTemperature_request(byte sensor_n)
{
  int packetSize;
    
    if (WiFiConnected && (RemoteMeasureState == STOP))
    {         
      Udp.begin(udpPort_base+sensor_n);
      packetSize = Udp.parsePacket();
      if (packetSize == 0)
      {
        Serial.println("Send a request to read remote temperature sensor ");
        Udp.beginPacket(udpAddress, udpPort_base+sensor_n);
        Udp.printf("READT");
        Udp.endPacket();
        Serial.print("READT command sent from: ");
        Serial.print(WiFi.localIP());
        Serial.print(", port ");
        Serial.println(udpPort_base+sensor_n);        
        retry_cnt = TX_RETRAY_N-1;
        RELOAD_TIMER(1);
        RemoteMeasureState = MEASURING;
      }
      else Udp.flush();       
    }  
    else
      RemoteMeasureState = STOP;
      
    return(RemoteMeasureState);
}

T_MT_STATES isRemoteTemperatureState(byte sensor_n)
{
  int  packetSize;
  char remote_temperature[10] ="";   
    
    if (RemoteMeasureState == MEASURING)  
    {    
      if (WiFiConnected)
      {
        packetSize = Udp.parsePacket();            
        if (packetSize > 0)
        {      
          Serial.print("Received temperature packet of size: ");
          Serial.print(packetSize);
      
          Serial.print(" From ");
          IPAddress remoteIp = Udp.remoteIP();
          Serial.print(remoteIp);
          Serial.print(", port ");
          Serial.print(Udp.remotePort());

          Udp.read(remote_temperature, 10);        
          Udp.flush();
        
          Serial.print(", contents is: ");
          Serial.println(remote_temperature);            

          Temperatures[dsDeviceCount+sensor_n] = atof(remote_temperature);          
          RemoteMeasureState = COMPLETED;
        }
        else if TIMEOUT_EVENT(1)
          if (retry_cnt > 0)
          {
            Udp.beginPacket(udpAddress,udpPort_base+sensor_n);
            Udp.printf("READT");
            Serial.println("Try again by reading remote temperature sensor");
            Serial.println("READT command sent");
            Udp.endPacket();
            RELOAD_TIMER(1);       
            retry_cnt--;                      
            RemoteMeasureState = MEASURING;
          } 
          else 
          {
            Serial.println("Remote sensor timeout..no more attempts to receive the remote temperature sensor measure");
            Temperatures[dsDeviceCount+sensor_n] = -125.00F;            
            RemoteMeasureState = COMPLETED;
          }          
      } // is WiFi connected
    }  // if (RemoteMeasureState == MEASURING)
             
    return(RemoteMeasureState);
}

T_MT_STATES RemoteTemperatureMeasurement_stop(void)
{
    // Pay attention, to test if the last measure has been completed use isRemoteTemperatureMeasurementState
    // then invoke RemoteTemperatureMeasurement_stop;
    RemoteMeasureState = STOP;
    return(RemoteMeasureState);
}

/* ------------------------------------------------------------  onCONNECTION ESTABILISHED ---------------------------------------------------------------------------- */

//wifi event handler
void WiFiEvent(WiFiEvent_t event)
{
    switch(event) 
    {
      case SYSTEM_EVENT_STA_GOT_IP:
        //When connected set 
        Serial.print("WiFi connected On event IP address: ");
        Serial.println(WiFi.localIP());  
        WiFiConnected = true;
      break;

      case SYSTEM_EVENT_STA_DISCONNECTED:
        Serial.println("WiFi lost connection");
        WiFiConnected = false;
        break;
      
      default: break;
    }    
}

void connectToWiFi(const char * ssid, const char * pwd) {
  Serial.println("Connecting to WiFi network: " + String(ssid));

  // delete old config
  WiFi.disconnect(true);

  //register event handler
  WiFi.onEvent(WiFiEvent);

  //Initiate connection
  WiFi.begin(ssid, pwd);
  Serial.println("Waiting for WIFI connection...");
}

 /* ------------------------------------------------------------  SETUP ------------------------------------------------------------------------------ */
 
void setup()
{
  int i;
  struct tm timeinfo;
      
    Serial.begin(115200);
    Serial.println("------- Multisensor temperature Acquisition System ----------");
    
    DayTimerSec = 0;
    
    Serial.println("DayTimerSec: " + String(DayTimerSec));
   
    pinMode(WIFI_STATUS, OUTPUT);
    digitalWrite(WIFI_STATUS, HIGH);
    
    Serial.println();

    ds.setOneWire(&dsWire);                                      // Assign oneWire pin for all devices
    ds.begin();                                                  // Start up the DS temperature sensor library    
    dsWire.reset_search();
    
    while (dsWire.search(ThermometersAdd[dsDeviceCount])) 
      dsDeviceCount++;
    
    Serial.print("Number of local temperature devices found...: ");
    Serial.println(dsDeviceCount);
        
    for (i=0; i<dsDeviceCount; i++)
    {
      printAddress(ThermometersAdd[i]);
      ds.setResolution(ThermometersAdd[i], ADC_TEMP_RESOLUTION); // Set ADC temperature resolution        
      Temperatures[i] = -126.00F;
    }

    for (i=0; i<N_REMOTE_TEMP_SENS; i++)          
      Temperatures[dsDeviceCount+i] = -126.00F;
        
    ds.setWaitForConversion(0);                   // 0 = ASYNC conversion activities ... don't wait the end of conversions       

    connectToWiFi(networkName, networkPswd);        
    WiFi.onEvent(WiFiEvent);
    
    SET_TIMEOUT(1, REMOTE_SENSOR_TIMEOUT_MS);
    RELOAD_TIMER(1);

    SET_TIMEOUT(2, LOCAL_SENSOR_TIMEOUT_MS);
    RELOAD_TIMER(2);

}
/*-------------------------------------------------------------------------------------------------------------------------------------------------- */
/* ------------------------------------------------------------  LOOP  ----------------------------------------------------------------------------- */
/*-------------------------------------------------------------------------------------------------------------------------------------------------- */
void loop()
{
  unsigned long currentMillis = millis();
               
 /* -------------------------------------------------- Sampling temperatures event -------------------------------------------------------*/      
     
    // On Sampling Temperature Event read temperure sensors, store them in Teperatures[], and print out them in the debug window
  
    if (SamplingEvent)
    {                 
      SamplingEvent = false;      
      Serial.println("Start measurements of the temperatures set...");      
      TemperatureMeasurement_start();                    
    }  

    if ((isTemperatureMeasurementState() == COMPLETED))    
    {    
      TemperatureMeasurement_stop();      
      if (EnableRemoteSensor)
        RemoteSensorTemperature_request(0);      
    }    

    if ((isRemoteTemperatureState(0) == COMPLETED) && EnableRemoteSensor)
    {           
      RemoteTemperatureMeasurement_stop();
      Temperatures_print();
    }
 /* -------------------------------------------------- Recording start/stop event -------------------------------------------------------*/
     
    if (WiFiConnected)
      digitalWrite(WIFI_STATUS, LOW);
    else
      digitalWrite(WIFI_STATUS, HIGH);  

 /* ----------------------------------------------------------------------------- -------------------------------------------------------*/

    if ((currentMillis-lastmils) > 1000)
    {
      lastmils = currentMillis;
      DayTimerSec++;
          
      if (!SamplingEvent)
        SamplingEvent = (DayTimerSec % TEMPERATURE_SAMPLING_TIME_S) == 0;

      if (SwTimer1 > 0) SwTimer1--;
      if (SwTimer2 > 0) SwTimer2--;
    }   
}
